
# from omnibelt import safe_self_execute

from typing import Iterator, Hashable

from collections import OrderedDict
from string import Formatter
from tqdm import tqdm
from tqdm.notebook import tqdm as tqdm_notebook


def format_readable_number(n, sig_figs):
	format_str = "{:." + str(sig_figs) + "g}"
	val = format_str.format(n)
	# remove trailing 0s
	if '.' in val:
		val = val.rstrip("0")
	# remove trailing .
	val = val.rstrip(".")
	return val


def human_readable_number(num, significant_figures=3, *, gap='', units=None):
	# Default units if not provided

	if num is None:
		return num

	if units is None:
		units = {
			"Q": 1_000_000_000_000_000,
			"T": 1_000_000_000_000,
			"B": 1_000_000_000,
			"M": 1_000_000,
			"K": 1_000,
			# "": 1
		}

	# Sort units from largest to smallest
	sorted_units = sorted(units.items(), key=lambda x: x[1], reverse=True)

	for unit, threshold in sorted_units:
		if abs(num) >= threshold:
			return format_readable_number(num / threshold, significant_figures) + gap + unit

	return format_readable_number(num, significant_figures)



def tqdmd(itr, key=None, **kwargs):
	pbar = tqdm(itr, **kwargs)
	for v in pbar:
		if key is not None:
			pbar.set_description(v if isinstance(key, bool) else key(v))
		yield v


def tqdmd_notebook(itr, key=None, **kwargs):
	pbar = tqdm_notebook(itr, **kwargs)
	for v in pbar:
		if key is not None:
			pbar.set_description(v if isinstance(key, bool) else key(v))
		yield v



def sign(x):
	return 0 if x == 0 else (1 if x > 0 else -1)


def expression_format(s, **vars):
	"""
	Evaluates the keys in the given string as expressions using the given variables
	"""
	fmt = Formatter()
	vals = {key:eval(key, vars) for _, key, _, _ in fmt.parse(s)}
	return s.format(**vals)



class PowerFormatter(Formatter):
	# TODO: partial formatting - only format fields that are specified, and leave others as is
	def get_field(self, field_name, args, kwargs):
		try:
			return super().get_field(field_name, args, kwargs)
		except: # TODO: find the right exception
			return eval(self.vformat(field_name, args, kwargs), kwargs), field_name
			# return f'{{{field_name}}}', field_name


	def parse(self, s):
		start_idx = -1
		escaped = ''
		pre_idx = 0
		counter = 0
		idx = 0

		while idx < len(s):
			open_idx = s.find("{", idx)
			close_idx = s.find("}", idx)

			if open_idx == -1 and close_idx == -1:
				if counter == 0:
					# raise StopIteration
					# print(f'ending with: {escaped + s[idx:]!r}')
					yield escaped + s[idx:], None, '', None
				else:
					raise ValueError("Mismatched '{' at index {}".format(start_idx))
				break

			if open_idx != -1 and (open_idx < close_idx or close_idx == -1):
				if counter == 0:
					# yield (s[idx:open_idx], None)
					start_idx = open_idx
					pre_idx = idx
				idx = open_idx + 1
				counter += 1

			if close_idx != -1 and (close_idx < open_idx or open_idx == -1):
				if counter == 0:
					raise ValueError("Mismatched '}' at index {}".format(close_idx))
				counter -= 1
				if counter == 0:
					pre = s[pre_idx:start_idx]
					field = s[start_idx + 1:close_idx]
					if field.startswith("{") and field.endswith("}"):
						escaped = pre + '{'

						for lit, field, spec, conv in self.parse(field[1:-1]):
							if escaped is not None:
								lit = escaped + lit
								escaped = None
							yield lit, field, spec, conv

						escaped = '}'

					else:
						# spec = None
						lim = field.rfind('}')
						conv_idx = field[lim+1:].find('!')
						if conv_idx != -1:
							conv = field[lim+2+conv_idx:]
							field = field[:lim+1+conv_idx]
						else:
							conv = None

						if conv is None:
							lim = field.rfind(']')
							spec_idx = field[lim+1:].find(':')
							if spec_idx != -1:
								spec = field[lim+2+spec_idx:]
								field = field[:lim+1+spec_idx]
							else:
								spec = ''
						else:
							spec_idx = conv.find(':')
							if spec_idx != -1:
								spec = conv[spec_idx+1:]
								conv = conv[:spec_idx]
							else:
								spec = ''

						# print(f'yielding: {escaped + pre!r}, {field!r}, {spec!r}, {conv!r}')
						# field = eval(self._format(field), self._world)
						yield escaped + pre, field, spec, conv
						escaped = ''
					start_idx = -1
				idx = close_idx + 1



def pformat(s, **vars):
	"""
	Evaluates the keys in the given string as expressions using the given variables (recursively)
	"""
	fmt = PowerFormatter()
	return fmt.format(s, **vars)




def safe_self_execute(obj, fn, default='<<short circuit>>',
				 flag='safe execute flag'):
	
	if flag in obj.__dict__:
		return default  # short circuit
	obj.__dict__[flag] = True
	
	try:
		out = fn()
	finally:
		del obj.__dict__['self printed flag']
	
	return out



def split_dict(items, keys):
	good, bad = OrderedDict(), OrderedDict()
	for k in items:
		if k in keys:
			good[k] = items[k]
		else:
			bad[k] = items[k]
	return good, bad



def filter_duplicates(*iterators: Iterator[Hashable]):
	seen = set()
	for itr in iterators:
		for x in itr:
			if x not in seen:
				seen.add(x)
				yield x










from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, IO
from pathlib import Path
from . import load_txt, save_txt, Packable, save_pack, load_pack, save_json, save_yaml, load_yaml, load_json
from .exporting import SimpleExporterBase
import dill
import toml


class UniversalExporter(SimpleExporterBase):
	@staticmethod
	def affinity_from_payload(payload: Any, **kwargs) -> bool:
		return True



class PickleExport(UniversalExporter, extensions='.pk'):
	@staticmethod
	def affinity_from_payload(payload: Any, **kwargs) -> bool:
		return dill.pickles(payload, **kwargs)

	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return dill.load(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		dill.dump(payload, path, **kwargs)



class PackedExport(UniversalExporter, extensions='.pkd'):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return load_pack(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_pack(payload, path, **kwargs)



class YamlExport(UniversalExporter, extensions=['.yml', '.yaml']):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return load_yaml(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_yaml(payload, path, **kwargs)



class TomlExport(UniversalExporter, extensions=['.toml', '.tml']):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return toml.load(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		toml.dump(payload, path.open('w'), **kwargs)



class JsonExport(SimpleExporterBase, types=[dict, list, str, int, float, bool, type(None)], extensions='.json'):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return load_json(path, **kwargs)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_json(payload, path, **kwargs)



class TextExport(SimpleExporterBase, types=str, extensions=['.txt', '.str']):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return load_txt(path)

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_txt(payload, path)



class IntExport(SimpleExporterBase, types=int, extensions='.int'):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return int(load_txt(path))

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_txt(str(payload), path)



class FloatExport(SimpleExporterBase, types=float, extensions='.float'):
	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		return float(load_txt(path))

	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		save_txt(str(payload), path)







from typing import Any, Iterator, List, Optional, Tuple, Union, Callable, Type, Dict
import inspect

from .typing import unspecified_argument
from .tricks import method_decorator
from .propagators import universal_propagator


class AbstractOperational:
	def __get__(self, instance, owner):
		if instance is None:
			return self
		return self._create_operator(instance, owner)


	def _create_operator(self, instance, owner):
		raise NotImplementedError



class AbstractOperator:
	def __init__(self, base: AbstractOperational, instance: Any, **kwargs):
		super().__init__(**kwargs)


	def _send_operation(self, item):
		raise NotImplementedError



class SimpleOperator(AbstractOperator):
	def __init__(self, base, instance, **kwargs):
		super().__init__(base, instance, **kwargs)
		self._base = base
		self._instance = instance

	class _operation_caller:
		def __init__(self, fn, instance, **kwargs):
			super().__init__(**kwargs)
			self.fn = fn
			self.instance = instance

		def __call__(self, *args, **kwargs):
			return self.fn(self.instance, *args, **kwargs)

	def _send_operation(self, fn: Callable, **kwargs):
		return self._operation_caller(fn, self._instance, **kwargs)



class AutoOperator(SimpleOperator):
	def _find_operation_target(self, item):
		return getattr(self._base, item)


	def _send_operation(self, item: str, **kwargs):
		return super()._send_operation(self._find_operation_target(item), **kwargs)



class FormatOperator(AutoOperator):
	def __init__(self, base, instance, formatter: str, **kwargs):
		super().__init__(base, instance, **kwargs)
		self._formatter = formatter


	def _find_operation_target(self, item):
		return getattr(self._base, self._formatter.format(item))



class PrefixOperator(AutoOperator):
	def __init__(self, base, instance, prefix: str, formatter: str = None, **kwargs):
		if formatter is None:
			formatter = f'{prefix}{"{}"}'
		super().__init__(base, instance, formatter=formatter, **kwargs)
		self._prefix = prefix



class KeyedOperator(AutoOperator):
	class _operation_caller(SimpleOperator._operation_caller):
		def __init__(self, fn, instance, *, key=None, **kwargs):
			super().__init__(fn, instance, **kwargs)
			self.key = key

		def __call__(self, *args, **kwargs):
			if self.key is not None:
				args = (self.key, *args)
			return super().__call__(*args, **kwargs)


	def _send_operation(self, item, *, key=unspecified_argument, **kwargs):
		if key is unspecified_argument:
			key = item
		return self._operation_caller(item, self._instance, key=key, **kwargs)



class HubOperator(KeyedOperator):
	def __init__(self, base, instance, *, hub_fn=None, **kwargs):
		if isinstance(hub_fn, str):
			hub_fn = getattr(base, hub_fn)
		super().__init__(base, instance, **kwargs)
		self._hub_fn = hub_fn


	def _find_operation_target(self, item):
		if self._hub_fn is None:
			return super()._find_operation_target(item)
		return self._hub_fn



class UniversalOperator(SimpleOperator):
	def _default_operation(self, item):
		return getattr(self._base, item)


	def __getattribute__(self, item):
		try:
			return super().__getattribute__(item)
		except AttributeError:
			return self._default_operation(item)



class MappedOperator(UniversalOperator):
	def __init__(self, base, instance, *, ops: Dict[str, Callable] = None, **kwargs):
		super().__init__(base, instance, **kwargs)
		self._ops = ops


	def _default_operation(self, item):
		if self._ops is None:
			return super()._default_operation(item)
		return self._send_operation(item)


	def _send_operation(self, fn: Union[Callable, str], **kwargs):
		if isinstance(fn, str):
			fn = self._ops[fn]
		return super()._send_operation(fn, **kwargs)


class ConditionalOperator(AutoOperator, UniversalOperator):
	def _check_if_operation(self, item):
		raise NotImplementedError


	def _default_operation(self, item):
		if self._check_if_operation(item):
			return self._send_operation(item)
		return super()._default_operation(item)



class OptionOperator(ConditionalOperator):
	def __init__(self, base, instance, *, ops=None, **kwargs):
		if ops is None:
			ops = set()
		super().__init__(base, instance, **kwargs)
		self._ops = ops


	def _check_if_operation(self, item):
		return item in self._ops



class AliasOperator(OptionOperator):
	def __init__(self, base, instance, *, aliases=None, **kwargs):
		if aliases is None:
			aliases = {}
		super().__init__(base, instance, **kwargs)
		self._ops = {**{op: op for op in self._ops}, **aliases}


	def _find_operation_target(self, item):
		return getattr(self._base, self._ops[item])



class _operation_base(method_decorator):
	def __init__(self, fn: Callable = None, **kwargs):
		super().__init__(fn, **kwargs)
		self._attr_name = None


	def _setup_decorator(self, owner: Type, name: str) -> 'method_decorator':
		self._attr_name = name
		return super()._setup_decorator(owner, name)


	@property
	def op_name(self):
		raise NotImplementedError


	@property
	def attr_name(self):
		return self._attr_name


	@property
	def original_function(self):
		return self._fn



class operation_base(_operation_base):
	# first argument of wrapped functions should always be the instance
	def __init__(self, fn: Union[str, Callable], **kwargs):
		if isinstance(fn, str):
			fn, name = None, fn
		else:
			name = None
		super().__init__(fn, **kwargs)
		self._op_name = name


	@property
	def op_name(self):
		if self._op_name is None:
			return self._attr_name
		return self._op_name



class auto_operation(_operation_base, universal_propagator):
	# first argument of wrapped functions should always be the instance
	def __call__(self, fn):
		raise ValueError('operation decorator does not take arguments')


	@property
	def op_name(self):
		if self._method_name is None:
			return self._attr_name
		return self._method_name



class SimpleOperational(AbstractOperational):
	Operator = AbstractOperator


	def _create_operator(self, instance, owner, **kwargs):
		return self.Operator(self, instance, **kwargs)



class OptionOperational(SimpleOperational):
	Operator = OptionOperator


	def operations(self) -> Iterator[str]:
		raise NotImplementedError


	def _create_operator(self, instance, owner, *, ops=None, **kwargs):
		if ops is None:
			ops = set(self.operations())
		return super()._create_operator(instance, owner, ops=ops, **kwargs)



class Operationalized(AbstractOperational):
	def as_operator(self, instance):
		return self._create_operator(instance, type(instance))



class DecoratedOperational(OptionOperational):
	Operator = MappedOperator

	_known_operations = None
	def __init_subclass__(cls, skip_inherit_operations=False, **kwargs):
		super().__init_subclass__(**kwargs)
		if '_known_operations' not in cls.__dict__:
			ops = {}
			if not skip_inherit_operations:
				for base in reversed(cls.__bases__):  # O-N
					new = getattr(base, '_known_operations', {})
					if new:
						ops.update(new)
			for name, attr in cls.__dict__.items(): # O-N
				if isinstance(attr, _operation_base):
					ops[attr.op_name] = attr
					setattr(cls, name, attr.original_function)
			cls._known_operations = ops # O-N

	def _process_operation(self, ops, name, attr):
		return op


	def operations(self) -> Iterator[str]:
		yield from self._known_operations.keys()


	def _create_operator(self, instance, owner, *, ops=None, **kwargs):
		if ops is None:
			ops = {name: getattr(self, op.attr_name)
			       for name, op in self._known_operations.items()}
		return super()._create_operator(instance, owner, ops=ops, **kwargs)







from omnibelt import pack, unpack, json_unpack, json_pack
from _util_test import get_adict


def test_json_pack_transactionable():
	'''Test transactionable functionality'''
	data = get_adict()
	
	data.q = 0
	assert 'q' in data and data.q == 0
	
	data.begin()

	assert 'q' in data and data.q == 0
	data.q = 1
	assert 'q' in data and data.q == 1
	
	data.commit()
	
	assert 'q' in data and data.q == 1
	
	data.begin()
	
	del data.q
	assert 'q' not in data
	
	data.abort()

	assert 'q' in data and data.q == 1



from typing import Tuple, List, Dict, Optional, Union, Any, Callable, Sequence, Iterator, Iterable, Type, Set


# from .abstract import
from .typing import agnostic


class AbstractCrafty:
	@classmethod
	def _emit_my_craft_items(cls, owner=None) -> Iterator[Tuple[str, 'AbstractCraft']]: # N-O
		if owner is None:
			owner = cls
		for key, val in reversed(cls.__dict__.items()): # N-O
			if isinstance(val, AbstractCraft):
				for craft in val.emit_craft_items(owner): # N-O
					yield key, craft



class AbstractSkill:
	def as_skill(self, owner: AbstractCrafty):
		return self



class AbstractCraft(AbstractSkill):
	def emit_craft_items(self, owner=None):
		yield self # stateless



class SkilledCraft(AbstractCraft):
	class Skill(AbstractSkill):
		def __init__(self, base: AbstractCraft, instance: AbstractCrafty, **kwargs):
			super().__init__(**kwargs)
			self._base = base
			self._instance = instance


	def as_skill(self, owner: AbstractCrafty):
		return self.Skill(self, owner) # stateful



class NestableCraft(AbstractCraft):
	def emit_craft_items(self, owner=None): # parsing order (N-O)
		yield from super().emit_craft_items(owner)
		content = self._wrapped_content()
		if isinstance(content, AbstractCraft):
			yield from content.emit_craft_items(owner)


	def _wrapped_content_leaf(self): # wrapped method
		wrapped = self._wrapped_content()
		return wrapped._wrapped_content_leaf() if isinstance(wrapped, NestableCraft) else wrapped


	def _wrapped_content(self): # wrapped method
		raise NotImplementedError



########################################################################################################################


class InheritableCrafty(AbstractCrafty):
	@agnostic
	def _emit_all_craft_items(self, *, remaining: Iterator[Type['InheritableCrafty']] = None,
	                          start : Type['InheritableCrafty'] = None, owner : Type['InheritableCrafty'] = None,
	                          **kwargs) -> Iterator[Tuple[Type[AbstractCrafty], str, AbstractCraft]]: # N-O
		cls = self if isinstance(self, type) else type(self)
		if remaining is None:
			remaining = iter(cls.mro()) # N-O
		if start is None:
			start = cls
		if owner is None:
			owner = self

		for current in remaining: # N-O
			if issubclass(current, AbstractCrafty):
				for key, craft in current._emit_my_craft_items(owner):
					yield current, key, craft
			if issubclass(current, InheritableCrafty):
				yield from current._emit_all_craft_items(remaining=remaining, start=start, owner=owner, **kwargs)



class ProcessedCrafty(InheritableCrafty):
	def _process_crafts(self):
		pass



class IndividualCrafty(ProcessedCrafty):
	def _process_crafts(self):
		for owner, key, craft in self._emit_all_craft_items():
			self._process_skill(owner, key, craft, craft.as_skill(self))


	def _process_skill(self, src: Type[AbstractCrafty], key: str, craft: AbstractCraft, skill: AbstractSkill):
		pass



class HiddenCrafty(AbstractCrafty):
	_hidden_crafts = None
	def __init_subclass__(cls, **kwargs):
		super().__init_subclass__(**kwargs)
		cls._hidden_crafts = []


	@classmethod
	def _emit_my_craft_items(cls, owner=None) -> Iterator[Tuple[str, 'AbstractCraft']]: # N-O
		yield from super()._emit_my_craft_items(owner)
		if cls._hidden_crafts is not None:
			yield from cls._hidden_crafts




########################################################################################################################
























from typing import Type, Optional, Union, Any, Callable, Sequence, Iterable, Iterator, Tuple, List, Dict, NamedTuple
from collections import OrderedDict
from functools import cached_property

from .tricks import method_decorator
from .typing import agnostic



class method_propagator(method_decorator):
	def __init__(self, *args, **kwargs):
		fn, args = self._filter_callable_arg(args)
		super().__init__(fn) # self._fn # decorated function
		self._method_name = None # of decorator:  @method_propagator.method_name(*args, **kwargs) \n def name(...)
		self._name = None # of decorated function
		self._args = args
		self._kwargs = kwargs


	_common_content_types = (cached_property, method_decorator, property, staticmethod, classmethod)

	def _filter_callable_arg(self, args):
		if len(args):
			first = args[0]
			if callable(first) or isinstance(first, self._common_content_types):
				return first, args[1:]
		return None, args


	def _setup_decorator(self, owner: Type, name: str):
		self._name = name
		return super()._setup_decorator(owner, name)


	@agnostic
	def _make_propagator(self, name, **kwargs): # subclasses should define methods which call this
		return self._propagator_reference(self, name, **kwargs)


	_propagation_type = None
	class _propagator_reference:
		'''Default propagator does not keep track of the originator'''
		def __init__(self, originator: Union['method_decorator', Type['method_decorator']], name, *,
		             propagation_type=None, **kwargs):
			if propagation_type is None:
				propagation_type = originator._propagation_type
				if propagation_type is None:
					propagation_type = originator if isinstance(originator, type) else type(originator)
			super().__init__(**kwargs)
			self.name = name
			self._propagator_type = propagation_type


		def __call__(self, *args, **kwargs):
			sub = self._propagator_type(*args, **kwargs)
			sub._method_name = self.name
			return sub



class _universal_propagator_meta(type):
	def __getattribute__(self, item):
		try:
			return super().__getattribute__(item)
		except AttributeError:
			return self._make_propagator(item)



class universal_propagator(method_propagator, metaclass=_universal_propagator_meta):
	def __getattribute__(self, item):
		try:
			return super().__getattribute__(item)
		except AttributeError:
			return self._make_propagator(item)



# class AbstractCollectorTrigger:
# 	@classmethod
# 	def process_collectors(cls, owner: Type['Collectable']):
# 		for key, val in owner.__dict__.items():
# 			if isinstance(val, method_propagator):
# 				setattr(owner, key, cls(owner, val))
#
#
# 	def __init__(self, owner: Type['Collectable'], base: method_propagator, **kwargs):
# 		super().__init__(**kwargs)



# class Collectable:
# 	def __init_subclass__(cls, collector_trigger: Optional[AbstractCollectorTrigger] = None, **kwargs):
# 		super().__init_subclass__(**kwargs)
# 		if collector_trigger is not None:
# 			collector_trigger.process_collectors(cls)
#
# 	def __init__(self, *args, collector_type=None, **kwargs):
# 		if collector_type is not None:
# 			collector_type.process_collectors(self)
# 		super().__init__(*args, **kwargs)



# class AbstractCollector:
# 	@classmethod
# 	def process_triggers(cls, source: Any, *,
# 	                     triggers: Optional[Iterator[Tuple[str, AbstractCollectorTrigger]]] = None):
# 		if triggers is None:
# 			triggers = ((name, trigger) for name, trigger in type(source).__dict__.items()
# 			            if isinstance(trigger, AbstractCollectorTrigger))
# 		for name, trigger in triggers:
# 			if isinstance(trigger, AbstractCollectorTrigger):
# 				setattr(source, name, cls(source, trigger))
#
#
# 	def __init__(self, source: Any, base: AbstractCollectorTrigger, **kwargs):
# 		super().__init__(**kwargs)


















from typing import List, Dict, Tuple, Optional, Union, Any, Hashable, Sequence, Callable, Generator, Type, Iterable, \
	Iterator, IO
from pathlib import Path
from itertools import chain
from collections import OrderedDict

from .typing import unspecified_argument
# from .registries import Class_Registry
from .logging import get_printer
prt = get_printer(__name__)



########################################################################################################################



class UnknownExportData(Exception):
	def __init__(self, obj):
		super().__init__(f'{obj}')
		self.obj = obj


class UnknownExportPath(Exception):
	def __init__(self, path):
		super().__init__(f'{str(path)}')
		self.path = path


class ExportFailedError(ValueError):
	def __init__(self, obj, fmts):
		super().__init__(f'{obj} failed using: {", ".join(fmts)}')
		self.fmts = fmts
		self.obj = obj


class LoadFailedError(Exception):
	def __init__(self, path, fmts):
		super().__init__(f'{path} failed using: {", ".join(fmts)}')
		self.path = path
		self.fmts = fmts


class AmbiguousLoadPathError(FileNotFoundError):
	def __init__(self, options):
		super().__init__(f'{options}')
		self.options = options



########################################################################################################################



class AbstractExportManager:
	def exporters(self) -> Iterator['AbstractExporter']: # N-O
		raise NotImplementedError


	def loaders(self) -> Iterator['AbstractLoader']:
		raise NotImplementedError


	def export(self, payload: Any, path: Union[str, Path], fmt: Union[str, 'AbstractExporter', None] = None, *,
			   root: Union[str, Path, None] = None, **kwargs) -> Path:
		raise NotImplementedError


	def load_export(self, path: Union[str, Path], fmt: Optional[Union[str, 'AbstractLoader']] = None, *,
	                root: Optional[Union[str, Path]] = None, **kwargs) -> Any:
		raise NotImplementedError



class AbstractLoader:
	def affinity_from_format(self, fmt: str) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given format.
		Non-positive values mean this exporter should not be used for this format.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		raise NotImplementedError


	def affinity_from_path(self, path: Path) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given path.
		Non-positive values mean this exporter should not be used for this path.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		return self.affinity_from_format(''.join(path.suffixes))


	@staticmethod
	def load_payload(src: AbstractExportManager, path: Path, *, fmt: Optional[str] = None, **kwargs) -> Any:
		raise NotImplementedError


	@classmethod
	def load_export(self, name: Optional[str] = None, root: Optional[Union[Path, str]] = None, *,
					path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None,
					fmt: Optional[Union[str, Type['ExporterBase']]] = unspecified_argument) -> Any:
		if manager is None:
			manager = getattr(self, '_my_export_manager', _current_export_manager)
		if fmt is unspecified_argument:
			fmt = self
		return manager.load_export(name=name, root=root, path=path, fmt=fmt)



class AbstractExporter:
	def affinity_from_format(self, fmt: str) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given format.
		Non-positive values mean this exporter should not be used for this format.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		raise NotImplementedError


	def affinity_from_path(self, path: Path) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given path.
		Non-positive values mean this exporter should not be used for this path.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		return self.affinity_from_format(''.join(path.suffixes))


	def affinity_from_payload(self, payload: Any) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given payload.
		Non-positive values mean this exporter should not be used for this payload.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		raise NotImplementedError


	# top level

	@classmethod
	def export(cls, payload, name: Optional[str] = None, root: Optional[Union[str, Path]] = None, *,
			   path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None,
			   fmt: Optional[Union[str, Type['ExporterBase']]] = unspecified_argument) -> Optional[Path]:
		if manager is None:
			manager = getattr(cls, '_my_export_manager', _current_export_manager)
		if fmt is unspecified_argument:
			fmt = cls
		return manager.export(payload, name=name, root=root, path=path, fmt=fmt)

	_LoadFailedError = LoadFailedError
	_ExportFailedError = ExportFailedError

	# custom workers

	@staticmethod
	def export_payload(src: AbstractExportManager, payload: Any, path: Path, *,
					   fmt: Optional[str] = None, **kwargs) -> Path:
		raise NotImplementedError



########################################################################################################################



class ExportManager(AbstractExportManager):
	def __init__(self, *, exporters: Optional[Iterable[AbstractExporter]] = None,
	             loaders: Optional[Iterable[AbstractLoader]] = None):
		self._exporters = list(exporters) if exporters is not None else []
		self._loaders = list(loaders) if loaders is not None else []


	def register(self, *exporters: Union[AbstractExporter, AbstractLoader,
										 Type[AbstractExporter], Type[AbstractLoader]]):
		for exporter in exporters:
			if (isinstance(exporter, type)
					and (issubclass(exporter, AbstractExporter) or issubclass(exporter, AbstractLoader))):
				exporter = exporter()
			if isinstance(exporter, AbstractExporter):
				self._exporters.append(exporter)
			elif isinstance(exporter, AbstractLoader):
				self._loaders.append(exporter)
			else:
				raise TypeError(f'Cannot register {exporter} as an exporter or loader')


	def exporters(self) -> Iterator[AbstractExporter]: # N-O
		yield from reversed(self._exporters)


	def _matching_exporters(self, fmt: Optional[str] = None, path: Optional[Path] = None,
							payload: Optional[Any] = unspecified_argument) -> Iterator[AbstractExporter]:
		assert fmt is not None or path is not None or payload is not unspecified_argument, \
			f'Must provide either a format, a path, or a payload: {fmt}, {path}, {payload}'
		options = list((None if payload is unspecified_argument else exporter.affinity_from_payload(payload),
					None if fmt is None else exporter.affinity_from_format(fmt),
					None if path is None else exporter.affinity_from_path(path),
					exporter) for exporter in self.exporters())

		for payload_affinity, fmt_affinity, path_affinity, exporter in sorted(options, reverse=True,
																			  key=lambda info: tuple(info[:-1])):
			if (payload_affinity is None or payload_affinity <= 0) \
				and (fmt_affinity is None or fmt_affinity <= 0) \
				and (path_affinity is None or path_affinity <= 0):
				break
			yield exporter


	def loaders(self) -> Iterator[AbstractExporter]: # N-O
		yield from reversed(self._loaders)


	def _matching_loaders(self, path: Path, fmt: Optional[str] = None) -> Iterator[AbstractLoader]:
		options = list((None if fmt is None else exporter.affinity_from_format(fmt),
					None if path is None else exporter.affinity_from_path(path),
					exporter) for exporter in self.exporters())

		for fmt_affinity, path_affinity, loader in sorted(options, reverse=True, key=lambda info: tuple(info[:-1])):
			if (fmt_affinity is None or fmt_affinity <= 0) and (path_affinity is None or path_affinity <= 0):
				break
			yield loader


	_UnknownExportData = UnknownExportData
	_ExportFailedError = ExportFailedError


	def export(self, payload: Any, path: Union[str, Path], *, fmt: Union[str, AbstractExporter, None] = None,
			   root: Union[str, Path, None] = None, **kwargs) -> Path:
		path = Path(path)
		if root is not None:
			path = Path(root) / path

		if isinstance(fmt, AbstractExporter):
			return fmt.export_payload(payload, path, src=self, **kwargs)

		errors = OrderedDict()
		for exporter in self._matching_exporters(payload=payload, fmt=fmt, path=path):
			try:
				return exporter.export_payload(self, payload=payload, path=path, fmt=fmt, **kwargs)
			except ExportFailedError as e:
				errors[exporter] = e

		if not errors:
			raise self._UnknownExportData(payload)
		raise self._ExportFailedError(errors)


	_UnknownLoadPath = UnknownExportPath
	_LoadFailedError = LoadFailedError


	def load_export(self, path: Union[str, Path], *, fmt: Optional[Union[str, AbstractLoader]] = None,
	                root: Optional[Union[str, Path]] = None, **kwargs) -> Any:
		path = Path(path)
		if root is not None:
			path = Path(root) / path

		if not path.exists(): # infer suffix only
			fallback = [p for p in path.parent.glob(f'{path.name}*') if p.stem == path.stem]
			if len(fallback) == 0:
				raise FileNotFoundError(path)
			elif len(fallback) > 1:
				raise AmbiguousLoadPathError(fallback)
			else:
				path = fallback[0]

		if isinstance(fmt, AbstractLoader):
			return fmt.load_payload(self, path, **kwargs)

		errors = OrderedDict()
		for loader in self._matching_loaders(path, fmt=fmt):
			try:
				return loader.load_payload(self, path=path, fmt=fmt, **kwargs)
			except LoadFailedError as e:
				errors[loader] = e

		if not errors:
			raise self._UnknownLoadPath(path)
		raise self._LoadFailedError(path, errors)



_current_export_manager = ExportManager()
def set_export_manager(manager: AbstractExportManager) -> AbstractExportManager:
	global _current_export_manager
	old = _current_export_manager
	_current_export_manager = manager
	return old



def export(obj: Any, path: Union[str, Path], *, fmt: Union[str, AbstractExporter, None] = None,
		   root: Union[str, Path, None] = None, manager: Optional[AbstractExportManager] = None,
		   **kwargs):
	if manager is None:
		manager = _current_export_manager
	return manager.export(obj, path=path, fmt=fmt, root=root, **kwargs)


def load_export(path: Union[str, Path], *, fmt: Union[str, AbstractLoader, None] = None,
				root: Union[str, Path, None] = None, manager: Optional[AbstractExportManager] = None,
				**kwargs):
	if manager is None:
		manager = _current_export_manager
	return manager.load_export(path=path, fmt=fmt, root=root, **kwargs)



class ExporterBase(AbstractExporter, AbstractLoader):
	_my_suffixes = None
	_my_payload_types = None
	def __init_subclass__(cls, extensions: Union[None, str, Sequence[str]] = None,
						  types: Union[None, Type, Sequence[Type]] = None, register: bool = None, **kwargs):
		super().__init_subclass__(**kwargs)
		if extensions is not None:
			extensions = extensions if isinstance(extensions, Sequence) else [extensions]
			extensions = [cls._parse_extension(ext) for ext in extensions]
			cls._my_suffixes = extensions

		if types is not None:
			types = types if isinstance(types, Sequence) else [types]
			cls._my_payload_types = tuple(types)

		if register is None:
			register = cls._my_suffixes is not None or cls._my_payload_types is not None
		if register:
			_current_export_manager.register(cls)


	@staticmethod
	def _parse_extension(ext):
		ext = ext.lower()
		if len(ext) and not ext.startswith('.'):
			ext = f'.{ext}'
		return ext


	def affinity_from_format(self, fmt: str) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given format.
		Non-positive values mean this exporter should not be used for this format.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		return self._my_suffixes is not None and self._parse_extension(fmt) in self._my_suffixes


	def affinity_from_path(self, path: Path) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given path.
		Non-positive values mean this exporter should not be used for this path.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		return self.affinity_from_format(''.join(path.suffixes))


	def affinity_from_payload(self, payload: Any) -> int:
		'''
		Returns a value representing whether this exporter should be used for the given payload.
		Non-positive values mean this exporter should not be used for this payload.
		The higher the value, the higher the priority for this exporter to be used.
		'''
		if self._my_payload_types is None or not isinstance(payload, self._my_payload_types):
			return False
		product = type(payload)
		history = product.mro()
		return len(history) - min(history.index(typ) for typ in self._my_payload_types if issubclass(product, typ))



class SimpleExporterBase(ExporterBase):
	def fix_path(self, path: Path) -> Path:
		if self._my_suffixes is not None and ''.join(path.suffixes) not in self._my_suffixes:
			path = path.with_suffix(self._my_suffixes[0])
		return path


	def export_payload(self, src: AbstractExportManager, payload: Any, path: Path, *,
					   fmt: Optional[str] = None, **kwargs) -> Path:
		if not len(path.suffix) and (fmt is None or self.affinity_from_format(fmt) > 0):
			# if no manual fmt was provided, or manually specified self
			path = self.fix_path(path.resolve())
		self._export_payload(payload, path, **kwargs)
		return path


	@staticmethod
	def _export_payload(payload: Any, path: Path, **kwargs) -> None:
		raise NotImplementedError


	def load_payload(self, src: AbstractExportManager, path: Path, *, fmt: Optional[str] = None, **kwargs) -> Any:
		if fmt is None or self.affinity_from_format(fmt) > 0:
			# if no manual fmt was provided, or manually specified self
			path = self.fix_path(path.resolve())
		if not path.exists():
			raise FileNotFoundError(path)
		return self._load_payload(path, **kwargs)


	@staticmethod
	def _load_payload(path: Path, **kwargs) -> Any:
		raise NotImplementedError





########################################################################################################################


#
#
# class OldExportManager:
#
# 	_export_fmts_head = []
# 	_export_fmts_tail = []
# 	_export_fmt_types: Dict[Type, List[Type[AbstractExporter]]] = OrderedDict()
# 	_export_fmt_exts: Dict[str, List[Type[AbstractExporter]]] = OrderedDict()
#
# 	# def __init_subclass__(cls, inherit_exporters=True, set_as_current=False, **kwargs):
# 	# 	super().__init_subclass__(**kwargs)
# 	#
# 	# 	head, tail = [], []
# 	# 	typs, exts = OrderedDict(), OrderedDict()
# 	#
# 	# 	if inherit_exporters:
# 	# 		for base in cls.__bases__:
# 	# 			if issubclass(base, ExportManager):
# 	# 				head.extend(base._export_fmts_head)
# 	# 				tail.extend(base._export_fmts_tail)
# 	# 				for k, v in base._export_fmt_types.items():
# 	# 					typs.setdefault(k, []).extend(v)
# 	# 				for k, v in base._export_fmt_exts.items():
# 	# 					exts.setdefault(k, []).extend(v)
# 	#
# 	# 	cls._export_fmts_head = head
# 	# 	cls._export_fmts_tail = tail
# 	# 	cls._export_fmt_types = typs
# 	# 	cls._export_fmt_exts = exts
# 	#
# 	# 	if set_as_current:
# 	# 		set_export_manager(cls)
#
# 	_UnknownExportData = UnknownExportData
# 	_UnknownExportPath = UnknownExportPath
# 	_UnknownExportFormat = UnknownExportFormat
#
# 	@classmethod
# 	def _related_fmts_by_type(cls, typ):
# 		options = [base for base in reversed(cls._export_fmt_types) if issubclass(typ, base)]
# 		history = list(typ.mro())
# 		for typ in sorted(options, key=lambda t: history.index(t)):
# 			yield from cls._export_fmt_types[typ]
#
# 	@classmethod
# 	def _related_fmts_by_path(cls, path):
# 		suffixes = path.suffixes
# 		if len(suffixes) == 0:
# 			suffixes = ['']
# 		for i in range(len(suffixes)):
# 			suffix = ''.join(suffixes[i:])
# 			if suffix in cls._export_fmt_exts:
# 				yield from cls._export_fmt_exts[suffix]
#
# 	@classmethod
# 	def resolve_fmt_from_obj(cls, obj: Any) -> Iterator[Type[AbstractExporter]]:
# 		missing = True
# 		for fmt in chain(reversed(cls._export_fmts_head),
# 		                 cls._related_fmts_by_type(type(obj)),
# 		                 cls._export_fmts_tail):
# 			if fmt.validate_export_obj(obj):
# 				missing = False
# 				yield fmt
#
# 		if missing:
# 			raise cls._UnknownExportData(obj)
#
# 	@classmethod
# 	def resolve_fmt_from_path(cls, path: Path) -> Iterator[Type[AbstractExporter]]:
# 		missing = True
# 		for fmt in chain(reversed(cls._export_fmts_head),
# 		                 cls._related_fmts_by_path(path), # TODO: maybe check for multi suffixes
# 		                 cls._export_fmts_tail):
# 			if fmt.validate_export_path(path):
# 				missing = False
# 				yield fmt
#
# 		if missing:
# 			raise cls._UnknownExportPath(path)
#
# 	@classmethod
# 	def resolve_fmt(cls, fmt: Union[str, Type, Type[AbstractExporter]]) -> Iterator[Type[AbstractExporter]]:
# 		if isinstance(fmt, type):
# 			if issubclass(fmt, Exporter):
# 				yield fmt
# 			else:
# 				yield from cls._related_fmts_by_type(fmt)
# 		else:
# 			assert isinstance(fmt, str), f'{fmt!r}'
# 			try:
# 				yield from cls.resolve_fmt_from_path(Path(f'null.{fmt}' if len(fmt) else 'null'))
# 			except cls._UnknownExportPath:
# 				raise cls._UnknownExportFormat(fmt)
#
# 	@classmethod
# 	def create_export_path(cls, name: str, *, root: Optional[Union[str, Path]] = None,
# 	                       fmt: Optional[str] = None) -> Path:
# 		if root is None:
# 			root = Path()
# 		root = Path(root)
#
# 		if fmt is None:
# 			return root / name
#
# 		for fmt in cls.resolve_fmt(fmt):
# 			return fmt._create_export_path(name, root, src=cls)
#
#
# 	@classmethod
# 	def create_load_path(cls, name: str, root: Optional[Union[str, Path]] = None):
# 		if root is None:
# 			root = Path()
# 		root = Path(root)
#
# 		if not root.exists():
# 			raise FileNotFoundError(root)
# 		options = list(root.glob(f'{name}*'))
# 		if not len(options):
# 			raise FileNotFoundError(root / name)
# 		if len(options) > 1:
# 			raise cls.AmbiguousLoadPathError(options)
# 		return options[0]
#
# 	@classmethod
# 	def _export_fmt(cls, fmt: Type['Exporter'], payload: Any, path: Path, **kwargs: Any) -> Path:
# 		return fmt._export_payload(payload, path=path, src=cls, **kwargs)
#
# 	@classmethod
# 	def _load_export_fmt(cls, fmt: Type['Exporter'], path: Path, **kwargs: Any) -> Path:
# 		return fmt._load_export(path, src=cls, **kwargs)
#
# 	@classmethod
# 	def export(cls, payload: Any, name: Optional[str] = None, root: Optional[Union[str, Path]] = None,
# 	           fmt: Optional[Union[str, Type, Type['Exporter']]] = None, path: Optional[Union[str, Path]] = None,
# 	           **kwargs) -> Path:
# 		assert path is not None or name is not None, f'Must provide either a path or a name to export: {payload}'
# 		if root is None and path is None and isinstance(name, Path):
# 			path = name
# 			name = None
# 		if root is not None:
# 			root = Path(root)
#
# 		if fmt is not None:
# 			fmts = cls.resolve_fmt(fmt)
# 		elif path is not None: # TODO: payload exporter has to figure out what to do if the extension is different
# 			fmts = cls.resolve_fmt_from_path(Path(path))
# 		else:
# 			fmts = cls.resolve_fmt_from_obj(payload)
#
# 		for fmt in fmts:
# 			dest = fmt.create_export_path(name=name, root=root, payload=payload) if path is None else Path(path)
# 			try:
# 				return cls._export_fmt(fmt, payload, dest, **kwargs)
# 			except ExportFailedError:
# 				pass
#
# 		raise cls.ExportFailedError(payload, fmts)
#
# 	@classmethod
# 	def load_export(cls, name: Optional[str] = None, root: Optional[Union[str, Path]] = None, *,
# 	                fmt: Optional[Union[str, Type['Exporter']]] = None, path: Optional[Union[str, Path]] = None,
# 	                **kwargs) -> Any:
# 		if root is None and path is None and isinstance(name, Path):
# 			path = name
# 			name = None
# 		if root is not None:
# 			root = Path(root)
#
# 		if fmt is not None:
# 			fmts = cls.resolve_fmt(fmt)
# 		elif path is not None:
# 			fmts = cls.resolve_fmt_from_path(path)
# 		else:
# 			path = cls.create_load_path(name=name, root=root)
# 			fmts = cls.resolve_fmt_from_path(path)
#
# 		for fmt in fmts:
# 			dest = fmt.create_export_path(name=name, root=root) if path is None else Path(path)
# 			try:
# 				return cls._load_export_fmt(fmt, dest, **kwargs)
# 			except fmt._LoadFailedError:
# 				pass
#
# 		raise cls.LoadFailedError(path, fmts)
#
# 	@classmethod
# 	def register(cls, exporter: Type['Exporter'], extensions: Optional[Union[str, Sequence[str]]] = None,
# 	             types: Optional[Union[Type, Sequence[Type]]] = None,
# 	             head: Optional[bool] = None, tail: Optional[bool] = None):
#
# 		if head is None and tail is None:
# 			head, tail = types is None, False
# 		if head:
# 			cls._export_fmts_head.append(exporter)
# 		if tail:
# 			cls._export_fmts_tail.append(exporter)
#
# 		if extensions is not None:
# 			if isinstance(extensions, str):
# 				extensions = (extensions,)
# 			else:
# 				extensions = tuple(extensions)
# 			for ext in extensions:
# 				if ext not in cls._export_fmt_exts:
# 					cls._export_fmt_exts[ext] = []
# 				cls._export_fmt_exts[ext].append(exporter)
#
# 		if types is not None:
# 			if isinstance(types, type):
# 				types = (types,)
# 			else:
# 				types = tuple(types)
# 			for typ in types:
# 				if typ not in cls._export_fmt_types:
# 					cls._export_fmt_types[typ] = []
# 				cls._export_fmt_types[typ].append(exporter)
#
# 		if extensions is None and types is None and not head and not tail:
# 			prt.warning(f'Exporter {exporter} is not registered to any extensions or types')
#


# class LoadFailedError(ValueError): pass
# class ExportFailedError(ValueError): pass


# class Exporter:
# 	@classmethod
# 	def validate_export_obj(cls, obj: Any) -> bool:
# 		options = getattr(cls, '_my_export_types', None)
# 		return options is not None and isinstance(obj, options)
#
# 	@classmethod
# 	def validate_export_path(cls, path: Path) -> bool:
# 		suffix = ''.join(path.suffixes)
# 		options = getattr(cls, '_my_export_extensions', None)
# 		return options is not None and len(suffix) and (suffix in options
# 		                                                or (suffix.startswith('.') and suffix[1:] in options))
#
#
# 	def __init_subclass__(cls, extensions: Optional[Union[str, Sequence[str]]] = None,
# 	                      types: Optional[Union[Type, Sequence[Type]]] = None,
# 			              head: Optional[bool] = None, tail: Optional[bool] = None,
# 	                      manager: Optional[ExportManager] = None, **kwargs):
# 		if extensions is not None:
# 			extensions = (extensions,) if isinstance(extensions, str) else tuple(extensions)
# 			extensions = tuple(ext if ext.startswith('.') else f'.{ext}' for ext in extensions)
# 		if types is not None:
# 			types = (types,) if isinstance(types, type) else tuple(types)
#
# 		_auto_manager = False
# 		if manager is None:
# 			_auto_manager = True
# 			manager = _current_export_manager
#
# 		super().__init_subclass__(**kwargs)
# 		manager.register(cls, extensions=extensions, types=types, head=head, tail=tail)
#
# 		if extensions is not None:
# 			cls._my_export_extensions = extensions
# 		if types is not None:
# 			cls._my_export_types = types
# 		if manager is not None and not _auto_manager:
# 			cls._my_export_manager = manager
#
#
# 	@classmethod
# 	def load_export(cls, name: Optional[str] = None, root: Optional[Union[Path, str]] = None, *,
# 	                path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None,
# 	                fmt: Optional[Union[str, Type['Exporter']]] = unspecified_argument) -> Any:
# 		if manager is None:
# 			manager = getattr(cls, '_my_export_manager', _current_export_manager)
# 		if fmt is unspecified_argument:
# 			fmt = cls
# 		return manager.load_export(name=name, root=root, path=path, fmt=fmt)
#
# 	@classmethod
# 	def export(cls, payload, name: Optional[str] = None, root: Optional[Union[str, Path]] = None, *,
# 	           path: Optional[Union[str, Path]] = None, manager: Optional['ExportManager'] = None,
# 	           fmt: Optional[Union[str, Type['Exporter']]] = unspecified_argument) -> Optional[Path]:
# 		if manager is None:
# 			manager = getattr(cls, '_my_export_manager', _current_export_manager)
# 		if fmt is unspecified_argument:
# 			fmt = cls
# 		return manager.export(payload, name=name, root=root, path=path, fmt=fmt)
#
#
# 	@classmethod
# 	def create_export_path(cls, name: str, root: Optional[Union[Path, str]], *,
# 	                       payload: Optional[Any] = unspecified_argument) -> Path:
# 		if root is None:
# 			root = Path()
# 		root = Path(root)
# 		options = getattr(cls, '_my_export_extensions', None)
# 		if options is not None and len(options) and not name.endswith(options[0]):
# 			name = f'{name}{options[0]}'
# 		return root / name
#
# 	_LoadFailedError = LoadFailedError
# 	_ExportFailedError = ExportFailedError
#
# 	@staticmethod
# 	def _load_export(path: Path, src: Type['ExportManager']) -> Any:
# 		raise NotImplementedError
#
# 	@staticmethod
# 	def _export_payload(payload: Any, path: Path, src: Type['ExportManager']) -> Optional[Path]:
# 		raise NotImplementedError


#
# class CollectiveExporter(Exporter):
# 	'''Usually braod exporters that can export multiple types of objects (eg. pickle, json, etc.)'''
# 	def __init_subclass__(cls, extensions=None, head=None, tail=None, **kwargs):
# 		super().__init_subclass__(extensions=extensions, head=head, tail=tail, **kwargs)
#
#
# class SelectiveExporter(Exporter):
# 	'''Usually specific exporters where the produced file is specific to the object type (eg. .png, .jpg, etc.)'''
# 	def __init_subclass__(cls, extensions=None, types=None, **kwargs):
# 		super().__init_subclass__(extensions=extensions, types=types, **kwargs)
#
#
# class Exportable(SelectiveExporter):
# 	'''Mixin for objects that can be exported with a custom export function :func:`_export_payload()`'''
# 	def __init_subclass__(cls, extensions=None, types=None, **kwargs):
# 		if types is None:
# 			types = cls
# 		super().__init_subclass__(extensions=extensions, types=types, **kwargs)
#
# 	def export(self, name: Optional[str] = None, root: Optional[Union[str, Path]] = None,
# 	           manager: Optional['ExportManager'] = None) -> Optional[Path]:
# 		return super().export(self, name=name, root=root, manager=manager)











from _util_test import get_adict
from omnibelt import adict, tdict, tlist, tset


def test_init():
	'''
	Test creating some standard humpack objects
	'''
	x = adict()
	assert len(x) == 0
	
	x = tlist()
	assert len(x) == 0
	
	x = tset()
	assert len(x) == 0


def test_adict():
	'''
	Test basic operations with tdict
	'''
	
	x = adict()
	assert str(x) == 't{}'
	
	x.a = 'a'
	assert str(x) == 't{a}'
	
	x.b = 10
	assert str(x) == 't{a, b}'
	assert len(x) == 2
	
	l = list(iter(x))
	assert str(l) == "['a', 'b']"
	
	assert x.a == 'a'
	assert x['a'] == 'a'


def test_elements():
	data = get_adict()
	
	data[dict] = data
	assert len(data) == len(data[dict])
	
	data[1234] = 'element'
	assert 1234 in data
	assert 'element' == data[1234]
	
	data.abc = 123
	assert 'abc' in data
	assert data.abc == 123
	assert data['abc'] == 123






# import numpy as np

from _util_test import get_adict
from omnibelt import json_pack, json_unpack
# from omnibelt.wrappers import Array

# def test_numpy_wrapper(): # TODO: transactionable arrays no longer supported
#
# 	x = Array(np.random.randn(4).astype(object))
#
# 	x[0] = get_adict()
# 	x[1] = 0
#
# 	assert x[1] == 0
#
# 	x.begin()
#
# 	x[0].arr = x
# 	x[1] = 10
#
# 	assert 'arr' in x[0]
# 	assert len(x[0].arr) == len(x)
# 	assert x[1] == 10
#
# 	assert repr(x) == repr(json_unpack(json_pack(x)))
#
# 	x.abort()
#
# 	assert x[1] == 0
# 	assert 'arr' not in x[0]
# 	assert repr(json_unpack(json_pack(x))) == repr(x)
	

# def test_numpy_arrays():
# 	x = np.random.randn(10)
# 	p = json_pack(x)
# 	c = json_unpack(p)
#
# 	assert (c-x).sum() == 0


# def test_complex_array(): # TODO: not currently supported on all systems
# 	x = np.arange(3).astype(object)
#
# 	x[0] = get_adict()
#
# 	p = json_pack(x)
# 	c = json_unpack(p)
#
# 	assert repr(x) == repr(c)





# from humpack import pack, unpack, json_unpack, json_pack
from omnibelt import pack, unpack, json_unpack, json_pack
from _util_test import get_adict


def test_pack_tdict():
	data = get_adict()
	rec = unpack(pack(data))
	
	assert repr(data) == repr(rec)


def test_json_pack_transactionable():
	data = get_adict()
	
	s = json_pack(data)
	assert repr(json_unpack(s)) == repr(data)
	
	data.begin()
	
	data[1234] = 'element'
	assert 1234 in data
	assert 'element' == data[1234]
	
	data.abc = 123
	assert 'abc' in data
	assert data.abc == 123
	assert data['abc'] == 123
	
	data.abort()
	
	assert repr(json_unpack(s)) == repr(data)
	
	